from typing import cast
from aiconsole.core.assets.load_all_assets import load_all_assets
from aiconsole.core.assets.agents.agent import Agent
from aiconsole.core.assets.asset import AssetType
from aiconsole.core.assets.materials.material import Material


async def get_all_agents() -> list[Agent]:
    lists = (await load_all_assets(AssetType.AGENT)).values()
    return cast(list[Agent], (list[0] for list in lists))


async def get_all_materials() -> list[Material]:
    lists = (await load_all_assets(AssetType.MATERIAL)).values()
    return cast(list[Material], (list[0] for list in lists))



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from aiconsole.core.chat.types import Chat
from aiconsole.core.chats.load_chat_history import load_chat_history
from aiconsole.core.chats.save_chat_history import save_chat_history
from aiconsole.core.project.paths import get_history_directory
from fastapi import APIRouter, Response, status
from fastapi.responses import JSONResponse
from send2trash import send2trash

router = APIRouter()


@router.delete("/{chat_id}")
async def delete_history(chat_id: str):
    file_path = get_history_directory() / f"{chat_id}.json"
    if file_path.exists():
        send2trash(file_path)
        return Response(
            status_code=status.HTTP_200_OK,
            content="Chat history deleted successfully",
        )
    else:
        return Response(
            status_code=status.HTTP_404_NOT_FOUND,
            content="Chat history not found",
        )


@router.get("/{chat_id}")
async def get_history(chat_id: str):
    chat = await load_chat_history(chat_id)

    return JSONResponse(chat.model_dump())


@router.patch("/{chat_id}")
async def save_history(chat_id, chat: Chat):
    if chat_id != chat.id:
        return Response(
            status_code=status.HTTP_400_BAD_REQUEST,
            content="Chat ID mismatch",
        )

    save_chat_history(chat)

    return Response(
        status_code=status.HTTP_201_CREATED,
        content="Chat history saved successfully",
    )


@router.get("/{chat_id}/path")
async def get_history_path(chat_id: str):
    return {"path": str(get_history_directory() / f"{chat_id}.json")}



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import litellm
from aiconsole.core.gpt.partial import GPTPartialResponse
from aiconsole.core.gpt.request import GPTRequest
from aiconsole.api.websockets.outgoing_messages import DebugJSONWSMessage
from .exceptions import NoOpenAPIKeyException
from .types import CLEAR_STR, GPTChoice, GPTResponse, GPTResponseMessage
from openai.error import AuthenticationError
from litellm.caching import Cache

_log = logging.getLogger(__name__)


litellm.cache = Cache()
litellm.cache.cache


class GPTExecutor:
    def __init__(self):
        self.request = {}
        self.response = GPTResponse(
            choices=[
                GPTChoice(
                    index=0,
                    message=GPTResponseMessage(role="assistant", content="Hello, how can I help you?"),
                    finnish_reason="",
                )
            ]
        )
        self.partial_response = GPTPartialResponse()

    async def execute(self, request: GPTRequest):
        request.validate_request()

        request_dict = {
            "max_tokens": request.max_tokens,
            "messages": request.get_messages_dump(),
            "model": request.model,
            "temperature": request.temperature,
            "presence_penalty": request.presence_penalty,
        }

        if request.tool_choice:
            request_dict["tool_choice"] = request.tool_choice

        if request.tools:
            request_dict["tools"] = [tool.model_dump() for tool in request.tools]

        for attempt in range(3):
            try:
                _log.info("Executing GPT request:", request_dict)
                self.request = request_dict
                response = await litellm.acompletion(
                    **request_dict,
                    stream=True,
                    # caching=True,
                )

                self.partial_response = GPTPartialResponse()

                async for chunk in response:  # type: ignore - for some reason response is not recognized as an async generator
                    self.partial_response.apply_chunk(chunk)
                    yield chunk
                    await asyncio.sleep(0)

                self.response = self.partial_response.to_final_response()

                if _log.isEnabledFor(logging.DEBUG):
                    await DebugJSONWSMessage(
                        message="GPT", object={"request": self.request, "response": self.response.model_dump()}
                    ).send_to_all()

                return
            except AuthenticationError:
                raise NoOpenAPIKeyException()
            except Exception as error:
                _log.exception(f"Error on attempt {attempt}: {error}", exc_info=error)
                if attempt == 2:
                    raise error
            _log.info("Retrying GPT request")
            yield CLEAR_STR

        raise Exception("Unable to complete GPT request.")



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from contextlib import asynccontextmanager
from logging import config

from aiconsole.api.routers import app_router
from aiconsole.consts import log_config
from aiconsole.core.project import project
from aiconsole.core.settings import project_settings
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware


@asynccontextmanager
async def lifespan(app: FastAPI):
    await project_settings.init()
    if project.is_project_initialized():
        await project.reinitialize_project()
    yield


config.dictConfig(log_config)


def app():
    origin = os.getenv("CORS_ORIGIN", None)

    if origin is None:
        raise Exception("CORS_ORIGIN environment variable not set")

    app = FastAPI(title="AI Console", lifespan=lifespan)

    app.add_middleware(
        CORSMiddleware,
        allow_origins=[origin],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    app.include_router(app_router)

    return app



import logging
import subprocess
import sys
from pathlib import Path

_log = logging.getLogger(__name__)


async def create_dedicated_venv():
    venv_path = get_current_project_venv_path()
    if not venv_path.exists():
        # should create the venv for the first time in the current project directory
        _log.info(f"Creating venv in {venv_path}")
        subprocess.Popen(
            [sys.executable, "-m", "venv", venv_path, "--system-site-packages"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        ).communicate()
    else:
        _log.info(f"Venv already exists in {venv_path}")


def get_current_project_venv_path():
    return Path.cwd() / ".aic" / "venv"


def get_current_project_venv_bin_path():
    if sys.platform == "win32":
        return get_current_project_venv_path() / "Scripts"
    else:
        return get_current_project_venv_path() / "bin"


def get_current_project_venv_python_path():
    return get_current_project_venv_bin_path() / "python"


def get_current_project_venv_available_packages():
    try:
        output = subprocess.check_output([str(get_current_project_venv_python_path() / "pip"), "list"]).decode("utf-8")
        package_lines = output.split("\\n")
        package_names = [line.split("==")[0] for line in package_lines]
        return " ".join(package_names)
    except subprocess.CalledProcessError:
        return ""



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from enum import Enum
import traceback

from aiconsole.core.assets.asset import Asset, AssetLocation, AssetType
from aiconsole.core.assets.asset import AssetStatus
from aiconsole.core.assets.materials.documentation_from_code import documentation_from_code
from aiconsole.core.assets.materials.rendered_material import RenderedMaterial
from typing import TYPE_CHECKING

from aiconsole.core.project.venv import (
    get_current_project_venv_available_packages,
    get_current_project_venv_python_path,
)

if TYPE_CHECKING:
    from aiconsole.core.assets.materials.content_evaluation_context import ContentEvaluationContext


class MaterialContentType(str, Enum):
    STATIC_TEXT = "static_text"
    DYNAMIC_TEXT = "dynamic_text"
    API = "api"


class Material(Asset):
    type: AssetType = AssetType.MATERIAL
    id: str
    name: str
    version: str = "0.0.1"
    usage: str
    defined_in: AssetLocation

    # Content, either static or dynamic
    content_type: MaterialContentType = MaterialContentType.STATIC_TEXT
    content_static_text: str = """

content, content content

## Sub header

Bullets in sub header:
* Bullet 1
* Bullet 2
* Bullet 3

""".strip()
    content_dynamic_text: str = """

import random
    
async def content(context):
    samples = ['sample 1' , 'sample 2', 'sample 3', 'sample 4']
    return f'''
# Examples of great content
{random.sample(samples, 2)}

'''.strip()

""".strip()
    content_api: str = """

'''
Add here general API description
'''

def create():
    '''
    Add comment when to use this function, and add example of usage:
    ```python
        create()
    ```
    '''
    print("Created")


def print_list():
    '''
    Use this function to print 'List'.
    Sample of use:
    ```python
        print_list()
    ```

    '''
    print("List")



def fibonacci(n):
    '''
    Use it to calculate and return the nth term of the Fibonacci sequence.
    Sample of use:
    ```python
      fibonacci(10)
    ```
    '''
    if n <= 0:
        return 0
    elif n == 1:
        return 1
    return fibonacci(n - 1) + fibonacci(n - 2)
""".strip()

    async def render(self, context: "ContentEvaluationContext"):
        header = f"# {self.name}\n\n"

        try:
            if self.content_type == MaterialContentType.DYNAMIC_TEXT:
                # Try compiling the python code and run it
                source_code = compile(self.content_dynamic_text, "<string>", "exec")
                # local vars should provide all the virtualenv specific stuff like:
                # - interpreter path
                # - installed packages
                local_vars = {
                    # this list should be updated with all needed variables from venv
                    "available_packages": get_current_project_venv_available_packages(),
                    "python_path": get_current_project_venv_python_path(),
                }
                # TODO add try/cache to catch potential plain text execution errors
                exec(source_code, local_vars)
                # currently, getting the python object from another interpreter is quite limited, and
                # using the dedicated local_vars is the easiest way (otherwise we would need to pickle
                # the object and send it to the other interpreter or use stdin/stdout)
                content_func = local_vars.get("content")
                if callable(content_func):
                    return RenderedMaterial(id=self.id, content=header + await content_func(context), error="")
                return RenderedMaterial(id=self.id, content="", error="No callable content function found!")
            elif self.content_type == MaterialContentType.STATIC_TEXT:
                return RenderedMaterial(id=self.id, content=header + self.content_static_text, error="")
            elif self.content_type == MaterialContentType.API:
                return RenderedMaterial(
                    id=self.id, content=header + documentation_from_code(self, self.content_api)(context), error=""
                )
        except Exception:
            return RenderedMaterial(id=self.id, content="", error=traceback.format_exc())

        raise ValueError("Material has no content")


class MaterialWithStatus(Material):
    status: AssetStatus = AssetStatus.ENABLED



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from aiconsole.core.assets.asset import AssetType
from pydantic import BaseModel


class BaseWSMessage(BaseModel):
    def get_type(self):
        return self.__class__.__name__

    def send_to_chat(self, chat_id: str):
        from aiconsole.api.websockets.connection_manager import send_message_to_chat

        return send_message_to_chat(chat_id, self)

    def send_to_all(self):
        from aiconsole.api.websockets.connection_manager import send_message_to_all

        return send_message_to_all(self)

    def model_dump(self):
        # Don't include None values, call to super to avoid recursion
        return {k: v for k, v in super().model_dump().items() if v is not None}


class NotificationWSMessage(BaseWSMessage):
    title: str
    message: str


class DebugJSONWSMessage(BaseWSMessage):
    message: str
    object: dict


class ErrorWSMessage(BaseWSMessage):
    error: str


class InitialProjectStatusWSMessage(BaseWSMessage):
    project_name: str | None = None
    project_path: str | None = None


class ProjectOpenedWSMessage(BaseWSMessage):
    name: str
    path: str


class ProjectClosedWSMessage(BaseWSMessage):
    pass


class ProjectLoadingWSMessage(BaseWSMessage):
    pass


class AssetsUpdatedWSMessage(BaseWSMessage):
    initial: bool
    asset_type: AssetType
    count: int


class SettingsWSMessage(BaseWSMessage):
    initial: bool


class RequestProcessingFinishedWSMessage(BaseWSMessage):
    request_id: str
    aborted: bool



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from aiconsole.core.assets.asset import AssetStatus
from aiconsole.core.project import project


import random


def create_materials_str() -> str:
    new_line = "\n"

    # We add forced becuase it may influence the choice of enabled materials
    available_materials = [
        *project.get_project_materials().assets_with_status(AssetStatus.FORCED),
        *project.get_project_materials().assets_with_status(AssetStatus.ENABLED),
    ]

    random_materials = (
        new_line.join([f"* {c.id} - {c.usage}" for c in random.sample(available_materials, len(available_materials))])
        if available_materials
        else ""
    )

    return random_materials



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pathlib import Path

from aiconsole.core.chats.load_chat_history import load_chat_history
from aiconsole.consts import MAX_RECENT_PROJECTS
from aiconsole.core.chats.list_possible_historic_chat_ids import list_possible_historic_chat_ids
from aiconsole.core.recent_projects.recent_project import RecentProject
from appdirs import user_config_dir


def _get_user_recent_projects_file():
    return Path(user_config_dir("aiconsole")) / "recent"


def _read_recent_projects():
    recent_projects_file = _get_user_recent_projects_file()
    if recent_projects_file.exists():
        recent_projects = [Path(path_str) for path_str in recent_projects_file.read_text().splitlines()]
    else:
        recent_projects = []

    return recent_projects


def _save_recent_projects(recent_projects: list[Path]):
    recent_projects_file = _get_user_recent_projects_file()
    recent_projects_file.parent.mkdir(parents=True, exist_ok=True)
    recent_projects_file.write_text("\n".join(str(p) for p in recent_projects))


async def add_to_recent_projects(project_path: Path):
    recent_projects = _read_recent_projects()
    recent_projects.insert(0, project_path)

    # only unique but keep order
    recent_projects = list(dict.fromkeys(recent_projects))

    # limit to MAX_RECENT_PROJECTS
    if len(recent_projects) > MAX_RECENT_PROJECTS:
        recent_projects = recent_projects[:MAX_RECENT_PROJECTS]

    _save_recent_projects(recent_projects)


async def remove_from_recent_projects(project_path: Path):
    recent_projects = _read_recent_projects()
    recent_projects.remove(project_path)
    _save_recent_projects(recent_projects)


async def get_recent_project():
    recent_projects = _read_recent_projects()

    recent_projects_real = []

    for path in recent_projects:
        recent_projects_real.append(
            RecentProject(
                name=os.path.basename(path),
                path=path,
                recent_chats=[
                    (await load_chat_history(id, path)).name for id in list_possible_historic_chat_ids(path)[:4]
                ],
            )
        )

    return recent_projects_real



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

import rtoml
from aiconsole.core.assets.agents.agent import Agent
from aiconsole.core.assets.asset import Asset, AssetLocation, AssetStatus, AssetType
from aiconsole.core.assets.materials.material import Material, MaterialContentType
from aiconsole.core.gpt.consts import GPTMode
from aiconsole.core.project.paths import get_core_assets_directory, get_project_assets_directory

_log = logging.getLogger(__name__)


async def load_asset_from_fs(asset_type: AssetType, asset_id: str, location: AssetLocation | None = None) -> Asset:
    """
    Load a specific asset.
    """

    project_dir_path = get_project_assets_directory(asset_type)
    core_resource_path = get_core_assets_directory(asset_type)

    if (project_dir_path / f"{asset_id}.toml").exists() and (
        location is None or location == AssetLocation.PROJECT_DIR
    ):
        if (core_resource_path / f"{asset_id}.toml").exists():
            _log.info(f"Asset {asset_id} exists in core and project directory")
        location = AssetLocation.PROJECT_DIR
        path = project_dir_path / f"{asset_id}.toml"
    elif (core_resource_path / f"{asset_id}.toml").exists() and (
        location is None or location == AssetLocation.AICONSOLE_CORE
    ):
        location = AssetLocation.AICONSOLE_CORE
        path = core_resource_path / f"{asset_id}.toml"
    else:
        raise KeyError(f"Asset {asset_id} not found")

    with open(path, "r") as file:
        tomldoc = rtoml.loads(file.read())

    asset_id = os.path.splitext(os.path.basename(path))[0]

    params = {
        "id": asset_id,
        "name": str(tomldoc.get("name", asset_id)).strip(),
        "version": str(tomldoc.get("version", "0.0.1")).strip(),
        "defined_in": location,
        "usage": str(tomldoc["usage"]).strip(),
        "usage_examples": tomldoc.get("usage_examples", []),
        "default_status": AssetStatus(str(tomldoc.get("default_status", "enabled")).strip()),
        "override": location == AssetLocation.PROJECT_DIR and (core_resource_path / f"{asset_id}.toml").exists(),
    }

    if asset_type == AssetType.MATERIAL:
        material = Material(
            **params,
            content_type=MaterialContentType(str(tomldoc["content_type"]).strip()),
        )

        if "content_static_text" in tomldoc:
            material.content_static_text = str(tomldoc["content_static_text"]).strip()

        if "content_dynamic_text" in tomldoc:
            material.content_dynamic_text = str(tomldoc["content_dynamic_text"]).strip()

        if "content_api" in tomldoc:
            material.content_api = str(tomldoc["content_api"]).strip()

        return material

    if asset_type == AssetType.AGENT:
        if asset_id == "user":
            raise KeyError("Agent 'user' is reserved.")

        params["system"] = str(tomldoc["system"]).strip()

        if "gpt_mode" in tomldoc:
            params["gpt_mode"] = GPTMode(str(tomldoc["gpt_mode"]).strip())

        if "execution_mode" in tomldoc:
            params["execution_mode"] = str(tomldoc["execution_mode"]).strip()

        agent = Agent(**params)

        return agent

    raise Exception(f"Asset type {asset_type} not supported.")



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib.util
import logging
import inspect
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from aiconsole.core.assets.materials.material import Material

_log = logging.getLogger(__name__)


def documentation_from_code(material: "Material", source: str):
    """
    Creates content of a material from a python material file.
    """

    def create_content(context):
        # Create a new module
        spec = importlib.util.spec_from_loader("temp_module", loader=None)

        if not spec:
            raise Exception("Could not create spec for temp_module")

        # Load the new module into memory
        python_module = importlib.util.module_from_spec(spec)

        # Compile the source into a code object that can be executed by exec()
        code_object = compile(source, "temp_module", "exec")

        # Define a new, blank namespace and execute the code object within it
        exec(code_object, python_module.__dict__)

        function_list = []
        for name, obj in inspect.getmembers(python_module):
            # take only locally defined exports, no imports
            if name.startswith("_"):
                continue

            if inspect.isfunction(obj):
                # Extract function signature
                async_prefix = "async " if inspect.iscoroutinefunction(obj) else ""
                signature = inspect.signature(obj)
                function_declaration = f"{async_prefix}def {name}{signature}"
                doc = inspect.getdoc(obj) or ""
                function_list.append(
                    f"""
{function_declaration}
{doc}
""".strip()
                    + "\n\n\n"
                )

        # get main docstring
        docstring = inspect.getdoc(python_module)

        newline = "\n"
        final_doc = f"""
{docstring + newline + newline if docstring else ''}

## Variables and Functions Available When Executing Python Code

{(newline + newline).join(function_list)}
""".strip()

        return final_doc

    return create_content



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from aiconsole.consts import DIRECTOR_MIN_TOKENS, DIRECTOR_PREFERRED_TOKENS
from aiconsole.core.chat.chat_outgoing_messages import SequenceStage, UpdateAnalysisWSMessage
from aiconsole.core.chat.types import Chat
from aiconsole.core.gpt.consts import GPTMode
from aiconsole.core.gpt.gpt_executor import GPTExecutor
from aiconsole.core.gpt.request import GPTRequest
from aiconsole.core.gpt.types import GPTRequestTextMessage
from aiconsole.utils.convert_messages import convert_messages


async def gpt_analysis_text_step(
    chat: Chat,
    request_id: str,
    gpt_mode: GPTMode,
    initial_system_prompt: str,
    last_system_prompt: str,
):
    gpt_executor = GPTExecutor()

    request = GPTRequest(
        system_message=initial_system_prompt,
        gpt_mode=gpt_mode,
        messages=[*convert_messages(chat), GPTRequestTextMessage(role="system", content=last_system_prompt)],
        presence_penalty=2,
        min_tokens=DIRECTOR_MIN_TOKENS,
        preferred_tokens=DIRECTOR_PREFERRED_TOKENS,
    )

    await UpdateAnalysisWSMessage(
        request_id=request_id,
        stage=SequenceStage.START,
    ).send_to_chat(chat.id)

    try:
        async for chunk in gpt_executor.execute(request):
            await UpdateAnalysisWSMessage(
                stage=SequenceStage.MIDDLE,
                request_id=request_id,
                agent_id=None,
                relevant_material_ids=None,
                next_step=None,
                thinking_process=gpt_executor.partial_response.choices[0].message.content,
            ).send_to_chat(chat.id)
    finally:
        await UpdateAnalysisWSMessage(
            request_id=request_id,
            stage=SequenceStage.END,
        ).send_to_chat(chat.id)

    return gpt_executor.response.choices[0].message.content or ""



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from aiconsole.core.assets.asset import AssetType
from aiconsole.core.project.paths import get_core_assets_directory, get_project_assets_directory
from aiconsole.core.project.project import is_project_initialized
from fastapi import APIRouter
from fastapi.responses import FileResponse

router = APIRouter()

log = logging.getLogger(__name__)


@router.get("/profile/{image}")
async def profile_image(image: str):
    if is_project_initialized():
        image_path = get_project_assets_directory(AssetType.AGENT) / image

        if image_path.exists():
            return FileResponse(str(image_path))

    static_path = get_core_assets_directory(AssetType.AGENT) / image
    if static_path.exists():
        return FileResponse(str(static_path))

    default_path = get_core_assets_directory(AssetType.AGENT) / "default.jpg"
    return FileResponse(str(default_path))



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any


def recursive_merge(dict1: dict[str, Any], dict2: dict[str, Any]) -> dict[str, Any]:
    """Recursively merge dictionaries."""
    for k, v in dict2.items():
        if k in dict1:
            if isinstance(dict1[k], dict) and isinstance(v, dict):
                recursive_merge(dict1[k], v)
            elif isinstance(dict1[k], list) and isinstance(v, list):
                dict1[k].extend(v)
                dict1[k] = list(set(dict1[k]))
            else:
                dict1[k] = v
        else:
            dict1[k] = v
    return dict1



from enum import Enum

from aiconsole.api.websockets.outgoing_messages import BaseWSMessage


class SequenceStage(str, Enum):
    START = "start"
    MIDDLE = "middle"
    END = "end"


class UpdateAnalysisWSMessage(BaseWSMessage):
    stage: SequenceStage
    request_id: str
    agent_id: str | None = None
    relevant_material_ids: list[str] | None = None
    next_step: str | None = None
    thinking_process: str | None = None


class UpdateMessageWSMessage(BaseWSMessage):
    request_id: str
    id: str
    stage: SequenceStage

    text_delta: str | None = None


class ResetMessageWSMessage(BaseWSMessage):
    request_id: str
    id: str


class UpdateToolCallWSMessage(BaseWSMessage):
    request_id: str
    id: str
    stage: SequenceStage

    language: str | None = None

    code_delta: str | None = None
    headline_delta: str | None = None


class UpdateToolCallOutputWSMessage(BaseWSMessage):
    request_id: str
    id: str
    stage: SequenceStage
    output_delta: str | None = None



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
    

def get_agent_system_message(agent):
    """
    Add common system message to all agent messages
    """

    return f"""
{agent.system}

When asked to display images, display them by outputing: ![image info](./pictures/image.png) in plain text, no function call required for displaying.

You are part of AIConsole multi-agent system that:
- write any text on an expert level.
- can perform any task.
- can execute code in user's local environment, and has full access to the file system.
- the executed code have full internet access.



""".strip()


from aiconsole.core.assets.asset import AssetLocation, AssetType
from aiconsole.core.project import project
from aiconsole.core.project.paths import get_project_assets_directory
from fastapi import HTTPException, Request
from fastapi.responses import JSONResponse


async def asset_exists(asset_type: AssetType, request: Request, asset_id: str):
    location_param = request.query_params.get("location", None)
    location = AssetLocation(location_param) if location_param else None

    if not location:
        raise HTTPException(status_code=400, detail="Location not specified")

    if asset_id == "new":
        return JSONResponse({"exists": False})
    else:
        if asset_type == AssetType.AGENT:
            assets = project.get_project_agents()
        elif asset_type == AssetType.MATERIAL:
            assets = project.get_project_materials()
        else:
            raise ValueError(f"Invalid asset type: {asset_type}")

        asset = assets.get_asset(asset_id, location)

        return JSONResponse({"exists": asset is not None})


def asset_path(asset_type: AssetType, request: Request, asset_id: str):
    if asset_type == AssetType.AGENT:
        asset = project.get_project_agents().get_asset(asset_id)
    elif asset_type == AssetType.MATERIAL:
        asset = project.get_project_materials().get_asset(asset_id)
    else:
        raise ValueError(f"Invalid asset type: {asset_type}")

    if asset is None:
        raise HTTPException(status_code=404, detail=f"Asset {asset_id} of type {asset_type} not found")

    path = get_project_assets_directory(asset_type) / f"{asset_id}.toml"

    return JSONResponse({"path": str(path)})



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from aiconsole.core.assets.asset import EditableObject

from aiconsole.core.code_running.code_interpreters.language_map import LanguageStr
from aiconsole.core.gpt.types import GPTRole
from pydantic import BaseModel


class AICToolCall(BaseModel):
    id: str
    language: LanguageStr
    code: str
    headline: str
    output: str | None = None


class AICMessage(BaseModel):
    id: str
    timestamp: str
    content: str
    tool_calls: list[AICToolCall]


class AICMessageGroup(BaseModel):
    id: str
    agent_id: str
    role: GPTRole
    task: str
    materials_ids: list[str]
    messages: list[AICMessage]


class ChatHeadline(EditableObject):
    last_modified: datetime

    def model_dump(self, **kwargs):
        return {
            **super().model_dump(**kwargs),
            "last_modified": self.last_modified.isoformat(),
        }


class Chat(ChatHeadline):
    title_edited: bool = False
    message_groups: list[AICMessageGroup]


class Command(BaseModel):
    command: str


class ChatHeadlines(BaseModel):
    headlines: list[ChatHeadline]



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from aiconsole.core.assets.asset import AssetStatus
from aiconsole.core.project import project


from openai_function_call import OpenAISchema
from pydantic import Field


import random


def create_plan_class(available_agents):
    enabled_materials = project.get_project_materials().assets_with_status(AssetStatus.ENABLED)

    class Plan(OpenAISchema):

        """
        Plan what should happen next.
        """

        thinking_process: str = Field(
            description="Short description of the thinking process that led to the next step.",
            json_schema_extra={"type": "string"},
        )

        next_step: str = Field(
            description="A short actionable description of the next single atomic task to move this conversation forward.",
            json_schema_extra={"type": "string"},
        )

        is_users_turn: bool = Field(
            ...,
            description="Whether the initiative is on the user side or on assistant side.",
            json_schema_extra={"type": "boolean"},
        )

        agent_id: str = Field(
            description="Chosen agent to perform the next step.",
            json_schema_extra={"enum": [s.id for s in random.sample(available_agents, len(available_agents))]},
        )

        relevant_material_ids: list[str] = Field(
            ...,
            description="Chosen material ids relevant for the task",
            json_schema_extra={
                "items": {
                    "enum": [k.id for k in random.sample(enabled_materials, len(enabled_materials))],
                    "type": "string",
                }
            },
        )

    return Plan



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import importlib
import logging
from typing import cast

from aiconsole.api.websockets.outgoing_messages import ErrorWSMessage, RequestProcessingFinishedWSMessage
from aiconsole.core.assets.agents.agent import Agent, ExecutionModeContext
from aiconsole.core.assets.materials.material import Material
from aiconsole.core.chat.types import Chat
from aiconsole.core.assets.materials.content_evaluation_context import ContentEvaluationContext
from aiconsole.core.project import project
from aiconsole.utils.cancel_on_disconnect import cancelable_endpoint
from fastapi import APIRouter, Request
from pydantic import BaseModel, ValidationError

router = APIRouter()
_log = logging.getLogger(__name__)


class ExecuteRequestData(BaseModel):
    request_id: str
    chat: Chat
    agent_id: str
    relevant_materials_ids: list[str]


async def dynamic_import_and_call_execution_mode(agent: Agent, context):
    execution_mode = agent.execution_mode

    split = execution_mode.split(":")

    if len(split) != 2:
        raise ValueError(
            f"Invalid execution_mode in agent {agent.name}: {execution_mode} - should be module_name:object_name"
        )

    module_name, object_name = execution_mode.split(":")
    module = importlib.import_module(module_name)
    obj = getattr(module, object_name, None)

    if obj is None:
        raise ValueError(f"Could not find {object_name} in {module_name} module in agent {agent.name}")

    if not callable(obj):
        raise ValueError(f"{object_name} in {module_name} is not callable (in agent {agent.name})")

    ret_val = obj(context)

    if not asyncio.iscoroutine(ret_val):
        raise ValueError(f"{object_name} in {module_name} is not a coroutine (in agent {agent.name})")

    await ret_val


@router.post("/{chat_id}/execute")
@cancelable_endpoint
async def execute(request: Request, data: ExecuteRequestData, chat_id):
    if chat_id != data.chat.id:
        raise ValidationError("Chat ID does not match")

    agent = cast(Agent, project.get_project_agents().get_asset(data.agent_id))

    content_context = ContentEvaluationContext(
        chat=data.chat,
        agent=agent,
        gpt_mode=agent.gpt_mode,
        relevant_materials=[
            cast(Material, project.get_project_materials().get_asset(id)) for id in data.relevant_materials_ids
        ],
    )

    rendered_materials = [await material.render(content_context) for material in content_context.relevant_materials]

    context = ExecutionModeContext(
        request_id=data.request_id,
        chat=data.chat,
        agent=agent,
        relevant_materials=rendered_materials,
        gpt_mode=agent.gpt_mode,
    )

    aborted = False
    try:
        await dynamic_import_and_call_execution_mode(agent, context)
    except asyncio.CancelledError:
        _log.warning("Cancelled execution mode")
        aborted = True
    except Exception as e:
        await ErrorWSMessage(error=str(e)).send_to_chat(data.chat.id)
        raise e
    finally:
        await RequestProcessingFinishedWSMessage(request_id=data.request_id, aborted=aborted).send_to_chat(
            data.chat.id
        )



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Literal
from pydantic import BaseModel


class GPTModel(str, Enum):
    GPT_4_0613 = "gpt-4-0613"
    GPT_35_TURBO_0613 = "gpt-3.5-turbo-0613"
    GPT_35_TURBO_16k_0613 = "gpt-3.5-turbo-16k-0613"
    GPT_4_11106_PREVIEW = "gpt-4-1106-preview"


class GPTMode(str, Enum):
    COST = "cost"
    SPEED = "speed"
    QUALITY = "quality"


class GPTEncoding(str, Enum):
    GPT_4 = "gpt-4"
    GPT_35 = "gpt-3.5-turbo"


GPTEncodingLiteral = Literal[GPTEncoding.GPT_4, GPTEncoding.GPT_35]


class GPTModelData(BaseModel):
    encoding: GPTEncodingLiteral
    max_tokens: int


MODEL_DATA: dict[str, GPTModelData] = {
    GPTModel.GPT_4_0613: GPTModelData(
        max_tokens=8192,
        encoding=GPTEncoding.GPT_4,
    ),
    GPTModel.GPT_35_TURBO_0613: GPTModelData(
        max_tokens=4096,
        encoding=GPTEncoding.GPT_35,
    ),
    GPTModel.GPT_35_TURBO_16k_0613: GPTModelData(
        max_tokens=16384,
        encoding=GPTEncoding.GPT_35,
    ),
    GPTModel.GPT_4_11106_PREVIEW: GPTModelData(
        max_tokens=128000,
        encoding=GPTEncoding.GPT_4,
    ),
}



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
from pathlib import Path
import uuid
from datetime import datetime

from aiconsole.core.chat.types import Chat

from aiconsole.core.project.paths import get_history_directory


async def load_chat_history(id: str, project_path: Path | None = None) -> Chat:
    history_directory = get_history_directory(project_path)
    file_path = history_directory / f"{id}.json"

    if file_path.exists():
        with open(file_path, "r") as f:
            data = json.load(f)

            # Convert old format
            if "message_groups" not in data or not data["message_groups"]:
                data["message_groups"] = []

                if "messages" in data and data["messages"]:
                    for message in data["messages"]:
                        data["message_groups"].append(
                            {
                                "id": message["id"] if "id" in message else uuid.uuid4().hex,
                                "role": message["role"] if "role" in message else "",
                                "task": message["task"] if "task" in message and message["task"] else "",
                                "agent_id": message["agent_id"] if "agent_id" in message else "",
                                "materials_ids": message["materials_ids"]
                                if "materials_ids" in message and message["materials_ids"]
                                else [],
                                "messages": [
                                    {
                                        "id": message["id"] if "id" in message else uuid.uuid4().hex,
                                        "timestamp": message["timestamp"] if "timestamp" in message else "",
                                        "content": message["content"] if "content" in message else "",
                                    }
                                ],
                            }
                        )
                    del data["messages"]

            # Add tool_calls to each message
            for group in data["message_groups"]:
                if "messages" in group and group["messages"]:
                    for msg in group["messages"]:
                        if "tool_calls" not in msg:
                            msg["tool_calls"] = []

            # For all tool calls without headline add an empty headline
            for group in data["message_groups"]:
                if "messages" in group and group["messages"]:
                    for msg in group["messages"]:
                        if "tool_calls" in msg and msg["tool_calls"]:
                            for tool_call in msg["tool_calls"]:
                                if "headline" not in tool_call:
                                    tool_call["headline"] = ""

            if "name" not in data or not data["name"]:
                if "headline" in data and data["headline"]:
                    data["name"] = data["headline"]
                elif "title" in data and data["title"]:
                    data["name"] = data["title"]
                else:

                    def extract_default_headline():
                        for group in data["message_groups"]:
                            if "messages" in group and group["messages"]:
                                for msg in group["messages"]:
                                    return msg.get("content")

                    data["name"] = extract_default_headline() or "New Chat"

            if "id" in data:
                del data["id"]

            if "last_modified" in data:
                del data["last_modified"]

            return Chat(id=id, last_modified=datetime.fromtimestamp(os.path.getmtime(file_path)), **data)
    else:
        return Chat(id=id, name="New Chat", title_edited=False, last_modified=datetime.now(), message_groups=[])



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import traceback
from uuid import uuid4

from aiconsole.api.websockets.outgoing_messages import ErrorWSMessage
from aiconsole.core.assets.agents.agent import ExecutionModeContext
from aiconsole.core.chat.chat_outgoing_messages import (
    SequenceStage,
    UpdateMessageWSMessage,
    UpdateToolCallOutputWSMessage,
    UpdateToolCallWSMessage,
)
from aiconsole.core.code_running.run_code import get_code_interpreter

_log = logging.getLogger(__name__)


async def execution_mode_example_countdown(
    context: ExecutionModeContext,
):
    last_message = "print('Hello world!')"

    message_id = str(uuid4())
    await UpdateMessageWSMessage(
        request_id=context.request_id,
        stage=SequenceStage.START,
        id=message_id,
        text_delta=f"This is a demo of execution mode. I will count down from 10 to 1 and then hello world code.\n\n",
    ).send_to_chat(context.chat.id)

    for i in range(10, 0, -1):
        await UpdateMessageWSMessage(
            request_id=context.request_id, stage=SequenceStage.MIDDLE, id=message_id, text_delta=f"{i}..."
        ).send_to_chat(context.chat.id)
        await asyncio.sleep(1)

    await UpdateMessageWSMessage(
        request_id=context.request_id,
        stage=SequenceStage.END,
        id=message_id,
    ).send_to_chat(context.chat.id)

    await asyncio.sleep(1)

    message_id = str(uuid4())
    await UpdateMessageWSMessage(
        request_id=context.request_id, stage=SequenceStage.START, id=message_id, text_delta=f"Done"
    ).send_to_chat(context.chat.id)

    await UpdateMessageWSMessage(
        request_id=context.request_id,
        stage=SequenceStage.END,
        id=message_id,
    ).send_to_chat(context.chat.id)

    tool_call_id = str(uuid4())

    await UpdateToolCallWSMessage(
        request_id=context.request_id,
        stage=SequenceStage.START,
        id=tool_call_id,
        code_delta=last_message,
        language="python",
    ).send_to_chat(context.chat.id)

    await UpdateToolCallWSMessage(
        request_id=context.request_id,
        stage=SequenceStage.END,
        id=tool_call_id,
    ).send_to_chat(context.chat.id)

    await UpdateToolCallOutputWSMessage(
        request_id=context.request_id, stage=SequenceStage.START, id=tool_call_id
    ).send_to_chat(context.chat.id)

    try:
        mats = context.relevant_materials

        try:
            async for token in get_code_interpreter("python").run(last_message, []):  # TODO add materials
                await UpdateToolCallOutputWSMessage(
                    request_id=context.request_id,
                    stage=SequenceStage.MIDDLE,
                    id=tool_call_id,
                    output_delta=token,
                ).send_to_chat(context.chat.id)
        except asyncio.CancelledError:
            get_code_interpreter("python").terminate()
            raise
        except Exception:
            await ErrorWSMessage(error=traceback.format_exc().strip()).send_to_chat(context.chat.id)
            await UpdateToolCallOutputWSMessage(
                request_id=context.request_id,
                stage=SequenceStage.MIDDLE,
                id=tool_call_id,
                output_delta=traceback.format_exc().strip(),
            ).send_to_chat(context.chat.id)
    except Exception as e:
        await ErrorWSMessage(error=str(e)).send_to_chat(context.chat.id)
        raise e
    finally:
        await UpdateToolCallOutputWSMessage(
            request_id=context.request_id, stage=SequenceStage.END, id=tool_call_id
        ).send_to_chat(context.chat.id)



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from uuid import uuid4

from aiconsole.core.assets.agents.agent import ExecutionModeContext
from aiconsole.core.chat.chat_outgoing_messages import (
    ResetMessageWSMessage,
    SequenceStage,
    UpdateMessageWSMessage,
    UpdateToolCallWSMessage,
)
from aiconsole.core.code_running.code_interpreters.language_map import language_map
from aiconsole.core.execution_modes.get_agent_system_message import get_agent_system_message
from aiconsole.core.gpt.create_full_prompt_with_materials import create_full_prompt_with_materials
from aiconsole.core.gpt.gpt_executor import GPTExecutor
from aiconsole.core.gpt.request import GPTRequest, ToolDefinition, ToolFunctionDefinition
from aiconsole.core.gpt.types import CLEAR_STR
from aiconsole.utils.convert_messages import convert_messages
from openai_function_call import OpenAISchema
from pydantic import BaseModel, Field

_log = logging.getLogger(__name__)


class CodeTask(OpenAISchema):
    headline: str = Field(
        ...,
        description="Short (max 25 chars) title of this task, it will be displayed to the user",
        json_schema_extra={"type": "string"},
    )


class python(CodeTask):
    """
    When you send a message containing Python code to python, it will be executed in a stateful Jupyter notebook environment
    """

    code: str = Field(
        ...,
        description="python code to execute, it will be executed in a stateful Jupyter notebook environment",
        json_schema_extra={"type": "string"},
    )


class shell(CodeTask):
    """
    This function executes the given code on the user's system using the local environment and returns the output.
    """

    code: str = Field(..., json_schema_extra={"type": "string"})


class applescript(CodeTask):
    """
    This function executes the given code on the user's system using the local environment and returns the output.
    """

    code: str = Field(..., json_schema_extra={"type": "string"})


async def execution_mode_interpreter(
    context: ExecutionModeContext,
):
    global llm

    system_message = create_full_prompt_with_materials(
        intro=get_agent_system_message(context.agent),
        materials=context.relevant_materials,
    )

    executor = GPTExecutor()

    class ToolCallStatus(BaseModel):
        id: str
        language: str | None = None
        code: str = ""
        headline: str = ""
        end_with_code: str = ""
        finished: bool | None = (
            False  # None is for marking that the finished status has been consumed and final message is sent
        )

    tool_calls_data: dict[str, ToolCallStatus] = {}

    async def finish_finished():
        for tool_call_id in tool_calls_data:
            tool_call_data = tool_calls_data[tool_call_id]
            if tool_call_data.finished:
                if tool_call_data.end_with_code:
                    await UpdateToolCallWSMessage(
                        request_id=context.request_id,
                        stage=SequenceStage.MIDDLE,
                        id=tool_call_id,
                        code_delta=tool_call_data.end_with_code,
                    ).send_to_chat(context.chat.id)
                await UpdateToolCallWSMessage(
                    request_id=context.request_id,
                    stage=SequenceStage.END,
                    id=tool_call_id,
                ).send_to_chat(context.chat.id)
                tool_call_data.finished = None

    message_id = str(uuid4())
    await UpdateMessageWSMessage(
        request_id=context.request_id,
        stage=SequenceStage.START,
        id=message_id,
    ).send_to_chat(context.chat.id)

    try:
        async for chunk in executor.execute(
            GPTRequest(
                system_message=system_message,
                gpt_mode=context.agent.gpt_mode,
                messages=[message for message in convert_messages(context.chat)],
                tools=[
                    ToolDefinition(type="function", function=ToolFunctionDefinition(**python.openai_schema)),
                    ToolDefinition(type="function", function=ToolFunctionDefinition(**shell.openai_schema)),
                    ToolDefinition(type="function", function=ToolFunctionDefinition(**applescript.openai_schema)),
                ],
                min_tokens=250,
                preferred_tokens=2000,
            )
        ):
            if chunk == CLEAR_STR:
                if message_id:
                    await ResetMessageWSMessage(
                        request_id=context.request_id,
                        id=message_id,
                    ).send_to_chat(context.chat.id)
                continue

            if "choices" not in chunk or len(chunk["choices"]) == 0:
                continue

            delta = chunk["choices"][0]["delta"]

            if "content" in delta and delta["content"]:
                await UpdateMessageWSMessage(
                    request_id=context.request_id,
                    stage=SequenceStage.MIDDLE,
                    id=message_id,
                    text_delta=delta["content"],
                ).send_to_chat(context.chat.id)

            tool_calls = executor.partial_response.choices[0].message.tool_calls

            for index, tool_call in enumerate(tool_calls):
                # All tool calls with lower indexes are finished
                if index > 0 and tool_calls_data[tool_calls[index - 1].id].finished is False:
                    tool_calls_data[tool_calls[index - 1].id].finished = True

                await finish_finished()

                if tool_call.id not in tool_calls_data:
                    tool_calls_data[tool_call.id] = ToolCallStatus(id=tool_call.id)
                    await UpdateToolCallWSMessage(
                        request_id=context.request_id,
                        stage=SequenceStage.START,
                        id=tool_call.id,
                    ).send_to_chat(context.chat.id)

                tool_call_data = tool_calls_data[tool_call.id]

                if tool_call.type == "function":
                    function_call = tool_call.function

                    async def send_language_if_needed(lang: str):
                        if tool_call_data.language is None:
                            tool_call_data.language = lang

                            await UpdateToolCallWSMessage(
                                request_id=context.request_id,
                                stage=SequenceStage.MIDDLE,
                                id=tool_call.id,
                                language=tool_call_data.language,
                            ).send_to_chat(context.chat.id)

                    async def send_code_delta(code_delta: str = "", headline_delta: str = ""):
                        if code_delta or headline_delta:
                            await UpdateToolCallWSMessage(
                                request_id=context.request_id,
                                stage=SequenceStage.MIDDLE,
                                id=tool_call.id,
                                code_delta=code_delta,
                                headline_delta=headline_delta,
                            ).send_to_chat(context.chat.id)

                    if function_call.arguments:
                        if function_call.name not in [python.__name__, shell.__name__, applescript.__name__]:
                            if tool_call_data.language is None:
                                await send_language_if_needed("python")

                                _log.info(f"function_call: {function_call}")
                                _log.info(f"function_call.arguments: {function_call.arguments}")

                                code_delta = f"{function_call.name}("  # TODO: This won't work
                                await send_code_delta(code_delta)
                                tool_call_data.end_with_code = ")"
                            else:
                                if isinstance(function_call.arguments, str):
                                    code_delta = function_call.arguments[len(tool_call_data.code) :]
                                    tool_call_data.code = function_call.arguments
                                    await send_code_delta(code_delta)
                        else:
                            arguments = function_call.arguments
                            languages = language_map.keys()

                            if tool_call_data.language is None and function_call.name in languages:
                                # Languge is in the name of the function call
                                await send_language_if_needed(function_call.name)

                            # This can now be both a string and a json object
                            if isinstance(arguments, str):
                                # We need to handle incorrect OpenAI responses, sometmes arguments is a string containing the code
                                if arguments and not arguments.startswith("{"):
                                    await send_language_if_needed("python")

                                    code_delta = arguments[len(tool_call_data.code) :]
                                    tool_call_data.code = arguments

                                    await send_code_delta(code_delta)
                            else:
                                code_delta = ""
                                headline_delta = ""

                                if arguments and "code" in arguments:
                                    await send_language_if_needed("python")

                                    code_delta = arguments["code"][len(tool_call_data.code) :]
                                    tool_call_data.code = arguments["code"]

                                if arguments and "headline" in arguments:
                                    headline_delta = arguments["headline"][len(tool_call_data.headline) :]
                                    tool_call_data.headline = arguments["headline"]

                                if code_delta or headline_delta:
                                    await send_code_delta(code_delta, headline_delta)
    finally:
        for tool_call_data in tool_calls_data.values():
            if tool_call_data.finished is False:
                tool_call_data.finished = True

        await finish_finished()

        await UpdateMessageWSMessage(
            request_id=context.request_id,
            stage=SequenceStage.END,
            id=message_id,
        ).send_to_chat(context.chat.id)



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from aiconsole.core.assets.materials.rendered_material import RenderedMaterial


def create_full_prompt_with_materials(intro: str, materials: list[RenderedMaterial], outro: str = ""):
    section_strs = []
    for material in materials:
        section_strs.append(material.content)

    # Construct the full prompt
    sub_str = "\n\n\n"
    full_prompt = f"{intro}\n\n\n{sub_str.join(section_strs)}\n\n\n{outro}".strip()

    return full_prompt



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from aiconsole.core.assets.asset import AssetType
from aiconsole.core.project import project
from aiconsole.core.settings.project_settings import get_aiconsole_settings
from fastapi import APIRouter
from fastapi.responses import JSONResponse

router = APIRouter()


@router.get("/")
async def agents_get():
    settings = get_aiconsole_settings()
    return JSONResponse(
        [
            *(
                {
                    **agent.model_dump(),
                    "status": settings.get_asset_status(AssetType.AGENT, agent.id),
                }
                for agent in project.get_project_agents().all_assets()
            )
        ]
    )



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from aiconsole.api.utils.asset_exists import asset_exists, asset_path
from aiconsole.api.utils.asset_get import asset_get
from aiconsole.api.utils.asset_save import asset_patch, asset_post
from aiconsole.api.utils.asset_status_change import asset_status_change
from aiconsole.api.utils.status_change_post_body import StatusChangePostBody
from aiconsole.core.assets.agents.agent import Agent, AgentWithStatus
from aiconsole.core.assets.asset import AssetLocation, AssetStatus, AssetType
from aiconsole.core.gpt.consts import GPTMode
from aiconsole.core.project import project
from fastapi import APIRouter, HTTPException, Request
from fastapi.responses import JSONResponse

router = APIRouter()


@router.get("/{agent_id}")
async def agent_get(request: Request, agent_id: str):
    return await asset_get(
        request,
        AssetType.AGENT,
        agent_id,
        lambda: AgentWithStatus(
            id="new_agent",
            name="New Agent",
            usage="",
            usage_examples=[],
            status=AssetStatus.ENABLED,
            defined_in=AssetLocation.PROJECT_DIR,
            gpt_mode=GPTMode.QUALITY,
            system="",
            override=False,
        ),
    )


@router.patch("/{agent_id}")
async def agent_patch(agent_id: str, agent: Agent):
    return await asset_patch(AssetType.AGENT, agent, agent_id)


@router.post("/{agent_id}")
async def agent_post(agent_id: str, agent: Agent):
    return await asset_post(AssetType.AGENT, agent, agent_id)


@router.post("/{agent_id}/status-change")
async def agent_status_change(agent_id: str, body: StatusChangePostBody):
    await asset_status_change(AssetType.AGENT, agent_id, body)


@router.delete("/{agent_id}")
async def delete_agent(agent_id: str):
    try:
        await project.get_project_agents().delete_asset(agent_id)
        return JSONResponse({"status": "ok"})
    except KeyError:
        raise HTTPException(status_code=404, detail="Agent not found")


@router.get("/{asset_id}/exists")
async def agent_exists(request: Request, asset_id: str):
    return await asset_exists(AssetType.AGENT, request, asset_id)


@router.get("/{asset_id}/path")
async def agent_path(request: Request, asset_id: str):
    return asset_path(AssetType.AGENT, request, asset_id)



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from aiconsole.core.recent_projects.recent_projects import get_recent_project, remove_from_recent_projects
from fastapi import APIRouter
from fastapi.responses import JSONResponse
from pydantic import BaseModel


class _DeleteProjectPayload(BaseModel):
    path: Path


router = APIRouter()


@router.get("/recent")
async def choose_project():
    return JSONResponse([rp.model_dump() for rp in await get_recent_project()])


@router.delete("/recent")
async def delete_project(data: _DeleteProjectPayload):
    await remove_from_recent_projects(data.path)



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from aiconsole.core.assets.asset import Asset, AssetStatus, AssetType
from pydantic import BaseModel
from aiconsole.core.chat.types import Chat

from aiconsole.core.gpt.consts import GPTMode
from aiconsole.core.assets.materials.rendered_material import RenderedMaterial


class Agent(Asset):
    type: AssetType = AssetType.AGENT
    system: str
    gpt_mode: GPTMode = GPTMode.QUALITY
    execution_mode: str = "aiconsole.core.execution_modes.normal:execution_mode_normal"


class AgentWithStatus(Agent):
    status: AssetStatus = AssetStatus.ENABLED


class ExecutionModeContext(BaseModel):
    request_id: str
    chat: Chat
    agent: Agent
    gpt_mode: GPTMode
    relevant_materials: list[RenderedMaterial]



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from aiconsole.core.chat.types import ChatHeadline
from aiconsole.api.endpoints.chats.chat import router
from aiconsole.core.chats.list_possible_historic_chat_ids import list_possible_historic_chat_ids
from aiconsole.core.chats.load_chat_history import load_chat_history

_log = logging.getLogger(__name__)


@router.get("/")
async def get_history_headlines():
    headlines = []

    for chat_id in list_possible_historic_chat_ids():
        try:
            chat = await load_chat_history(chat_id)

            if chat:
                headlines.append(
                    ChatHeadline(id=chat_id, name=chat.name, last_modified=chat.last_modified).model_dump()
                )

        except Exception as e:
            _log.exception(e)
            _log.error(f"Failed to get history: {e} {chat_id}")

    return headlines



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Literal
from pydantic import BaseModel
from typing_extensions import TypedDict


CLEAR_STR = "<<<< CLEAR >>>>"
CLEAR_STR_TYPE = Literal["<<<< CLEAR >>>>"]


GPTRole = Literal["user", "assistant", "system", "tool"]


class EnforcedFunctionCallFuncSpec(TypedDict):
    name: str


class EnforcedFunctionCall(TypedDict):
    type: Literal["function"]
    function: EnforcedFunctionCallFuncSpec


class GPTFunctionCall(BaseModel):
    name: str
    arguments: dict | str


class GPTToolCall(BaseModel):
    id: str
    type: str = "function"
    function: GPTFunctionCall


class GPTResponseMessage(BaseModel):
    role: GPTRole
    content: str | None = None
    tool_calls: list[GPTToolCall] = []
    name: str | None = None

    def model_dump(self):
        # Don't include None values, call to super to avoid recursion
        return {k: v for k, v in super().model_dump().items() if v is not None}


class GPTRequestToolMessage(BaseModel):
    role: GPTRole = "tool"
    content: str | None
    tool_call_id: str


class GPTRequestTextMessage(BaseModel):
    role: GPTRole
    content: str | None = None
    name: str | None = None
    tool_calls: list[GPTToolCall] | None = None

    def model_dump(self):
        # Don't include None values, call to super to avoid recursion
        return {k: v for k, v in super().model_dump().items() if v is not None}


GPTRequestMessage = GPTRequestTextMessage | GPTRequestToolMessage


class GPTChoice(BaseModel):
    index: int
    message: GPTResponseMessage
    finnish_reason: str


class GPTResponse(BaseModel):
    id: str = ""
    object: str = ""
    created: int = 0
    model: str = ""
    choices: list[GPTChoice]



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import logging
from logging import config
from contextlib import asynccontextmanager
import os
from fastapi import FastAPI
from aiconsole.core.project import project
from aiconsole.core.settings import project_settings
from aiconsole.consts import log_config
from uvicorn import run


@asynccontextmanager
async def lifespan(app: FastAPI):
    await project_settings.init()
    if project.is_project_initialized():
        await project.reinitialize_project()
    yield


config.dictConfig(log_config)

_log = logging.getLogger(__name__)


def run_aiconsole(dev: bool):
    parser = argparse.ArgumentParser(description="Start the backend server.")
    parser.add_argument("--port", type=int, help="Port to listen on.", default=8000)
    parser.add_argument("--origin", type=str, help="Origin for the frontend.", default="http://localhost:3000")

    port = parser.parse_args().port
    origin = parser.parse_args().origin
    os.environ["CORS_ORIGIN"] = origin

    try:
        run(
            "aiconsole.app:app",
            host="0.0.0.0",
            port=port,
            reload=dev,
            factory=True,
        )
    except KeyboardInterrupt:
        _log.info("Exiting ...")


def aiconsole_dev():
    run_aiconsole(dev=True)


def aiconsole():
    run_aiconsole(dev=False)


if __name__ == "__main__":
    aiconsole_dev()



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
from aiconsole.api.websockets.outgoing_messages import RequestProcessingFinishedWSMessage

from fastapi import APIRouter, HTTPException, Request

from aiconsole.core.analysis.director import director_analyse
from aiconsole.core.chat.types import Chat
from aiconsole.utils.cancel_on_disconnect import cancelable_endpoint
from pydantic import BaseModel, ValidationError

router = APIRouter()
_log = logging.getLogger(__name__)


class AnalysisRequestData(BaseModel):
    request_id: str
    chat: Chat


@router.post("/{chat_id}/analyse")
@cancelable_endpoint
async def analyse(request: Request, data: AnalysisRequestData, chat_id):
    if chat_id != data.chat.id:
        raise ValidationError("Chat ID does not match")

    aborted = False
    try:
        return await director_analyse(data.chat, data.request_id)
    except asyncio.CancelledError:
        _log.info("Analysis cancelled")
        aborted = True
    except Exception as e:
        _log.exception("Analysis failed", exc_info=e)
        raise HTTPException(status_code=400, detail=str(e))
    finally:
        await RequestProcessingFinishedWSMessage(request_id=data.request_id, aborted=aborted).send_to_chat(
            data.chat.id
        )



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
from uuid import uuid4
from aiconsole.core.chat.chat_outgoing_messages import ResetMessageWSMessage, SequenceStage, UpdateMessageWSMessage
from aiconsole.core.assets.agents.agent import ExecutionModeContext
from aiconsole.core.execution_modes.get_agent_system_message import get_agent_system_message
from aiconsole.core.gpt.gpt_executor import GPTExecutor
from aiconsole.utils.convert_messages import convert_messages
from aiconsole.core.gpt.create_full_prompt_with_materials import create_full_prompt_with_materials
from aiconsole.core.gpt.types import CLEAR_STR
from aiconsole.core.gpt.request import GPTRequest

_log = logging.getLogger(__name__)


async def execution_mode_normal(
    context: ExecutionModeContext,
):
    _log.debug("execution_mode_normal")

    gpt_executor = GPTExecutor()

    message_id = str(uuid4())
    await UpdateMessageWSMessage(
        request_id=context.request_id,
        stage=SequenceStage.START,
        id=message_id,
    ).send_to_chat(context.chat.id)

    try:
        async for chunk in gpt_executor.execute(
            GPTRequest(
                messages=convert_messages(context.chat),
                gpt_mode=context.gpt_mode,
                system_message=create_full_prompt_with_materials(
                    intro=get_agent_system_message(context.agent),
                    materials=context.relevant_materials,
                ),
                min_tokens=250,
                preferred_tokens=2000,
            )
        ):
            if chunk == CLEAR_STR:
                if message_id:
                    await ResetMessageWSMessage(
                        request_id=context.request_id,
                        id=message_id,
                    ).send_to_chat(context.chat.id)
                continue

            await UpdateMessageWSMessage(
                request_id=context.request_id,
                stage=SequenceStage.MIDDLE,
                id=message_id,
                text_delta=chunk["choices"][0]["delta"].get("content", ""),
            ).send_to_chat(context.chat.id)
    finally:
        await UpdateMessageWSMessage(
            request_id=context.request_id,
            stage=SequenceStage.END,
            id=message_id,
        ).send_to_chat(context.chat.id)
